/*------------------------------------------------------
 Linear deterministic fractals
 Linearni deterministicke fraktaly
 
 Cantor discontinium
 Koch snowflake
 Sierpinski triangle
 Sierpinski quad
 
 OpenSceneGraph default camera manipulation with mouse
 
 SPACE                 = viewall
 S (repeately clicked) = scenegraph info
 KEY UP                = incr recursion
 KEY DOWN              = decr recursion
 
 (C) 2009 Jan Stalmach, http://stalmach.wz.cz 
 Contact at stalmachjan@gmail.com
 ENG/CZ, OpenSceneGraph 2.8
 -------------------------------------------------------*/

#include <osgViewer/ViewerEventHandlers>

/// vertical shift for cantor discontinium (just for draw)
#define POSUN 0.05

/// handy type
typedef GLfloat bod[2];

/// global scene root
osg::ref_ptr<osg::Group> root = new osg::Group;


/** Fractal class **/
class Fractals : public osg::Referenced
{
  public:
  Fractals(int iterace);
  void Generate();
  void setIterace(int iter) {m_iterace = iter;};
  int  getIterace() {return m_iterace;};
  
  void Cantor(bod begin, bod end, int n);
  void Sierpinski(bod A, bod B, bod C, int n);
  void Sierpinski_ctverec(bod A, bod B, bod C, bod D, int n);
  void Kochova_vlocka(bod point1, bod point2, bod point3);
  
  osg::ref_ptr<osg::Group> m_cantorNode;
  osg::ref_ptr<osg::Group> m_kochovaNode;
  osg::ref_ptr<osg::Group> m_sierpinskiTriNode;
  osg::ref_ptr<osg::Group> m_sierpinskiQuadNode;
  
  private:
  void Init();
  void Kochova(bod begin, bod end, int n);
  
  osg::ref_ptr<osg::Geode> m_lineKoch;
  osg::ref_ptr<osg::Geode> m_lineCantor;
  osg::ref_ptr<osg::Geode> m_lineSTri;
  osg::ref_ptr<osg::Geode> m_lineSQuad;
  
  osg::ref_ptr<osg::Geometry>  m_lineGeometryKoch;
  osg::ref_ptr<osg::Vec4Array> m_lineColorKoch;
  osg::ref_ptr<osg::Vec3Array> m_lineVertexKoch;
  
  osg::ref_ptr<osg::Geometry>  m_lineGeometryCantor;
  osg::ref_ptr<osg::Vec4Array> m_lineColorCantor;
  osg::ref_ptr<osg::Vec3Array> m_lineVertexCantor;  
  
  osg::ref_ptr<osg::Geometry>  m_lineGeometrySTri;
  osg::ref_ptr<osg::Vec4Array> m_lineColorSTri;
  osg::ref_ptr<osg::Vec3Array> m_lineVertexSTri;  
  
  osg::ref_ptr<osg::Geometry>  m_lineGeometrySQuad;
  osg::ref_ptr<osg::Vec4Array> m_lineColorSQuad;
  osg::ref_ptr<osg::Vec3Array> m_lineVertexSQuad;  
  
  int m_iterace;
};


/** constructor **/
Fractals::Fractals(int iterace) : m_iterace(iterace)
{
  Generate();
}


/** Initialize data **/
void Fractals::Init()
{
  m_cantorNode          = new osg::Group;
  m_kochovaNode         = new osg::Group;
  m_sierpinskiTriNode   = new osg::Group;
  m_sierpinskiQuadNode  = new osg::Group;
  
  m_lineCantor = new osg::Geode;
  m_lineKoch   = new osg::Geode;
  m_lineSTri   = new osg::Geode;
  m_lineSQuad  = new osg::Geode;
  
  m_lineGeometryKoch = new osg::Geometry;
  m_lineColorKoch    = new osg::Vec4Array;
  m_lineVertexKoch   = new osg::Vec3Array;
  
  m_lineColorKoch->push_back(osg::Vec4f(1.0, 1.0, 0.0, 1.0));
  m_lineGeometryKoch->setColorArray(m_lineColorKoch.get());
  m_lineGeometryKoch->setVertexArray(m_lineVertexKoch.get());
  m_lineGeometryKoch->setColorBinding(osg::Geometry::BIND_OVERALL);
  
  m_lineKoch->addDrawable(m_lineGeometryKoch.get());
  m_kochovaNode->addChild(m_lineKoch.get());
  
  m_lineGeometryCantor = new osg::Geometry;
  m_lineColorCantor    = new osg::Vec4Array;
  m_lineVertexCantor   = new osg::Vec3Array;
  
  m_lineColorCantor->push_back(osg::Vec4f(1.0, 1.0, 1.0, 1.0));
  m_lineGeometryCantor->setColorArray(m_lineColorCantor.get());
  m_lineGeometryCantor->setVertexArray(m_lineVertexCantor.get());
  m_lineGeometryCantor->setColorBinding(osg::Geometry::BIND_OVERALL);
  
  m_lineCantor->addDrawable(m_lineGeometryCantor.get());
  m_cantorNode->addChild(m_lineCantor.get());
  
  m_lineGeometrySTri = new osg::Geometry;
  m_lineColorSTri    = new osg::Vec4Array;
  m_lineVertexSTri   = new osg::Vec3Array;
  
  m_lineColorSTri->push_back(osg::Vec4f(1.0, 0.0, 0.0, 1.0));
  m_lineGeometrySTri->setColorArray(m_lineColorSTri.get());
  m_lineGeometrySTri->setVertexArray(m_lineVertexSTri .get());
  m_lineGeometrySTri->setColorBinding(osg::Geometry::BIND_OVERALL);
  
  m_lineSTri->addDrawable(m_lineGeometrySTri.get());
  m_sierpinskiTriNode->addChild(m_lineSTri.get());
  
  m_lineGeometrySQuad = new osg::Geometry;
  m_lineColorSQuad    = new osg::Vec4Array;
  m_lineVertexSQuad   = new osg::Vec3Array;
  
  m_lineColorSQuad->push_back(osg::Vec4f(1.0, 0.0, 1.0, 1.0));
  m_lineGeometrySQuad->setColorArray(m_lineColorSQuad.get());
  m_lineGeometrySQuad->setVertexArray(m_lineVertexSQuad.get());
  m_lineGeometrySQuad->setColorBinding(osg::Geometry::BIND_OVERALL);
  
  m_lineSQuad->addDrawable(m_lineGeometrySQuad.get());
  m_sierpinskiQuadNode->addChild(m_lineSQuad.get());
}


/** Cantor discontinium **/
void Fractals::Cantor(bod begin, bod end, int n)
{
	if(n>=0)   
	{   
		// zacatek
		bod pointA = {begin[0],begin[1]-POSUN};  
		// 2/3   
		bod pointB = {(2.0 * begin[0]+end[0])/3.0,((2.0 * begin[1]+end[1])/3.0)-POSUN};       
		// 1/3 delky    
		bod pointC = {(begin[0]+2.0 * end[0])/3.0,((begin[1]+2.0 *end[1])/3.0)-POSUN};    
		// koncovy bod	
		bod pointD = {end[0],end[1]-POSUN}; 
		
		Cantor(pointA, pointB,n-1);
		Cantor(pointC, pointD,n-1);
    
    m_lineVertexCantor->push_back(osg::Vec3f(begin[0],0.f,begin[1]));
    m_lineVertexCantor->push_back(osg::Vec3f(end[0],0.f,end[1]));             
	}
}


/** Koch snowflake one side **/
void Fractals::Kochova(bod begin, bod end, int n)
{
	if(n<=0)  
	{    
    m_lineVertexKoch->push_back(osg::Vec3f(begin[0],0.f,begin[1]));
    m_lineVertexKoch->push_back(osg::Vec3f(end[0],0.f,end[1])); 
	}  
  else  
	{     
		// begin
		bod pointA = {begin[0],begin[1]};  
		// 2/3   
		bod pointB = {(2.0 * begin[0]+end[0])/3.0,(2.0 * begin[1]+end[1])/3.0 };   
		// 1/2    
		bod pointC = {(begin[0]+end[0])/2.0 - sqrt(3.0)/6.0 *(end[1]-begin[1]),
			(begin[1]+end[1])/2.0 + sqrt(3.0)/6.0 *(end[0]-begin[0])};    
		// 1/3    
		bod pointD = {(begin[0]+2.0 * end[0])/3.0,(begin[1]+2.0 *end[1])/3.0 };    
		// end	
		bod pointE = {end[0],end[1]}; 
		
		// a -> b  
		Kochova(pointA, pointB, n-1);   	  
		// b -> c  
		Kochova(pointB, pointC, n-1);   
		// c -> d  
		Kochova(pointC, pointD, n-1);   
		// d -> e    
		Kochova(pointD, pointE, n-1);   
		
	}
}


/** Koch snowflake **/
void Fractals::Kochova_vlocka(bod point1, bod point2, bod point3)
{     
	Kochova(point1, point2, m_iterace);   
	Kochova(point2, point3, m_iterace);   
	Kochova(point3, point1, m_iterace);
} 


/** Sierpinskeho trojuhelnik **/
void Fractals::Sierpinski(bod A, bod B, bod C, int n)
{
	if(n<=0)  
	{    
    m_lineVertexSTri->push_back(osg::Vec3f(A[0],0.f,A[1]));
    m_lineVertexSTri->push_back(osg::Vec3f(B[0],0.f,B[1]));
    m_lineVertexSTri->push_back(osg::Vec3f(B[0],0.f,B[1])); 
    m_lineVertexSTri->push_back(osg::Vec3f(C[0],0.f,C[1]));
    m_lineVertexSTri->push_back(osg::Vec3f(C[0],0.f,C[1]));
    m_lineVertexSTri->push_back(osg::Vec3f(A[0],0.f,A[1]));
	}  
  else  
	{     
		bod pointA = {A[0],A[1]};
		bod pointB = {(A[0]+B[0])/2.0 ,(A[1]+B[1])/2.0 };
		bod pointC = {(A[0]+C[0])/2.0 ,(A[1]+C[1])/2.0 }; 
		bod pointD = {(B[0]+C[0])/2.0 ,(B[1]+C[1])/2.0 };
		bod pointE = {B[0],B[1]};
		bod pointF = {C[0],C[1]}; 
		
		Sierpinski(pointA, pointB, pointC, n-1); 
		Sierpinski(pointB, pointE, pointD, n-1);
		Sierpinski(pointC, pointD, pointF, n-1); 
	}
}


/** Sierpinskeho ctverec **/
void Fractals::Sierpinski_ctverec(bod A, bod B, bod C, bod D, int n)
{
	if(n<=0)  
	{    
    m_lineVertexSQuad->push_back(osg::Vec3f(A[0],0.f,A[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(B[0],0.f,B[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(B[0],0.f,B[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(C[0],0.f,C[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(C[0],0.f,C[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(D[0],0.f,D[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(D[0],0.f,D[1]));
    m_lineVertexSQuad->push_back(osg::Vec3f(A[0],0.f,A[1]));
	}  
  else  
	{     
		bod pointA = {A[0],A[1]};
		bod pointB = {(A[0]*2.0+B[0])/3.0 ,(A[1]*2.0+B[1])/3.0 };
		bod pointC = {(A[0]*2.0+C[0])/3.0 ,(A[1]*2.0+C[1])/3.0 }; 
		bod pointD = {(A[0]*2.0+D[0])/3.0 ,(A[1]*2.0+D[1])/3.0 };
        
		bod pointE = {(A[0]+B[0]*2.0)/3.0 ,(A[1]+B[1]*2.0)/3.0 };
		bod pointF = {(B[0]*2.0+D[0])/3.0 ,(B[1]*2.0+D[1])/3.0 }; 
        
		bod pointG = {B[0],B[1]};
		bod pointH = {(B[0]*2.0+C[0])/3.0 ,(B[1]*2.0+C[1])/3.0 };
        
		bod pointI = {(D[0]*2.0+B[0])/3.0 ,(D[1]*2.0+B[1])/3.0 };
		bod pointJ = {(A[0]+D[0]*2.0)/3.0 ,(A[1]+D[1]*2.0)/3.0 };
        
		bod pointK = {(C[0]*2.0+B[0])/3.0 ,(C[1]*2.0+B[1])/3.0 };
		bod pointL = {(A[0]+C[0]*2.0)/3.0 ,(A[1]+C[1]*2.0)/3.0 };
        
		bod pointM = {(D[0]*2.0+C[0])/3.0 ,(D[1]*2.0+C[1])/3.0 };
		bod pointN = {D[0],D[1]};
        
		bod pointO = {(C[0]*2.0+D[0])/3.0 ,(C[1]*2.0+D[1])/3.0 };
        
		bod pointP = {C[0],C[1]};
        
		Sierpinski_ctverec(pointA, pointB, pointC, pointD, n-1);
		Sierpinski_ctverec(pointB, pointE, pointF, pointC, n-1);
		Sierpinski_ctverec(pointE, pointG, pointH, pointF, n-1);
		Sierpinski_ctverec(pointD, pointC, pointI, pointJ, n-1);
		Sierpinski_ctverec(pointF, pointH, pointK, pointL, n-1);
		Sierpinski_ctverec(pointJ, pointI, pointM, pointN, n-1);
		Sierpinski_ctverec(pointI, pointL, pointO, pointM, n-1);
		Sierpinski_ctverec(pointL, pointK, pointP, pointO, n-1);
	}
}


/** create scenegraph **/
void Fractals::Generate()
{
  Init();
      
  // Kochova vlocka
	bod A = {-1,0};
	bod B = {0,1.44};
	bod C = {1,0};
	Kochova_vlocka(A,B,C);
  
  osg::ref_ptr<osg::DrawElementsUInt> index = new osg::DrawElementsUInt(osg::PrimitiveSet::LINE_STRIP, 0);
  for(unsigned int i = 0; i<m_lineVertexKoch->size(); i++)
  {
    index->push_back(i);
  }
  
  m_lineGeometryKoch->addPrimitiveSet(index.get());
    
	// Cantorovo diskontinium
	bod X = {-0.5,0.5};
	bod Y = {0.5,0.5};
	Cantor(X,Y, m_iterace);
  
  osg::ref_ptr<osg::DrawElementsUInt> indexCantor = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
  for(unsigned int i = 0; i<m_lineVertexCantor->size(); i++)
  {
    indexCantor->push_back(i);
  }
  
  m_lineGeometryCantor->addPrimitiveSet(indexCantor.get());
    
	// Sierpinskeho trojuhelnik
	bod P = {0,1.25};
	bod Q = {-0.5,0.55};
	bod R = {0.5,0.55};
	Sierpinski(P,Q,R, m_iterace);
  
  osg::ref_ptr<osg::DrawElementsUInt> indexSTri = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
  for(unsigned int i = 0; i<m_lineVertexSTri->size(); i++)
  {
    indexSTri->push_back(i);
  }
  
  m_lineGeometrySTri->addPrimitiveSet(indexSTri.get());
    
	// Sierpinskeho ctverec
	bod I = {-0.2,0.2};
	bod J = {0.2,0.2};
	bod K = {0.2,-0.2};
	bod L = {-0.2,-0.2};
	Sierpinski_ctverec(I,J,K,L, m_iterace);
  
  osg::ref_ptr<osg::DrawElementsUInt> indexSQuad = new osg::DrawElementsUInt(osg::PrimitiveSet::LINES, 0);
  for(unsigned int i = 0; i<m_lineVertexSQuad->size(); i++)
  {
    indexSQuad->push_back(i);
  }
  
  m_lineGeometrySQuad->addPrimitiveSet(indexSQuad.get());
}


/** keybord manipulator **/
class KeyboardEventHandler : public osgGA::GUIEventHandler
{
public:
  
  KeyboardEventHandler(Fractals* fractals):
  _fractals(fractals) {}
  
  virtual bool handle(const osgGA::GUIEventAdapter& ea,osgGA::GUIActionAdapter&)
  {
    switch(ea.getEventType())
    {
      case(osgGA::GUIEventAdapter::KEYDOWN):
      {
        if (ea.getKey() == osgGA::GUIEventAdapter::KEY_Up)
        {
          _fractals->setIterace(_fractals->getIterace()+1);
          _fractals->Generate();
          root->removeChild(0, 4);
          root->addChild(_fractals->m_cantorNode.get());
          root->addChild(_fractals->m_kochovaNode.get());
          root->addChild(_fractals->m_sierpinskiTriNode.get());
          root->addChild(_fractals->m_sierpinskiQuadNode.get());
        }
        
        if (ea.getKey() == osgGA::GUIEventAdapter::KEY_Down && _fractals->getIterace()>0)
        {
          _fractals->setIterace(_fractals->getIterace()-1);
          _fractals->Generate();
          root->removeChild(0, 4);
          root->addChild(_fractals->m_cantorNode.get());
          root->addChild(_fractals->m_kochovaNode.get());
          root->addChild(_fractals->m_sierpinskiTriNode.get());
          root->addChild(_fractals->m_sierpinskiQuadNode.get());
        }
        
        return false;
      }
        
      default:
        return false;
    }
  }
  
  osg::ref_ptr<Fractals> _fractals;
  
};


/** app entry **/
int main()
{
  
  osg::ref_ptr<Fractals> F = new Fractals(1);
    
  osgViewer::Viewer viewer;
  osg::DisplaySettings::instance()->setNumMultiSamples(8);
  
  viewer.addEventHandler(new osgViewer::StatsHandler);
  viewer.addEventHandler(new KeyboardEventHandler(F.get()));
    
  osg::ref_ptr<osg::StateSet> ss = root->getOrCreateStateSet();
  ss->setMode(GL_LIGHTING,osg::StateAttribute::OFF);
    
  root->addChild(F->m_cantorNode.get());
  root->addChild(F->m_kochovaNode.get());
  root->addChild(F->m_sierpinskiTriNode.get());
  root->addChild(F->m_sierpinskiQuadNode.get());

  viewer.setSceneData(root.get());

  return viewer.run();
}
